/*
 * Renderer 6. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;

/**
   Draw a single line on a 45 degree diagonal to the camera's
   near plane. Watch the interpolated vertex and color values
   as you move the near plane forwards and backwards across
   this line.
*/
public class TestNearPlaneClipping_v2 implements KeyListener
{
   private Scene scene;      // The event handlers need
   private final JFrame jf;  // access to these fields.
   private final FrameBufferPanel fbp;
   private double near = 1.0;
   private boolean perspective = false;
   private boolean showCamera = true;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public TestNearPlaneClipping_v2()
   {
      scene = new Scene("TestNearPlaneClipping_v2",
                        Camera.projOrtho());
      final Model model = new Model("LineDiagonalToNearPlane");
      final Position position = new Position(model);
      scene.addPosition(position);

      // Two Vertex objects.
      model.addVertex(new Vertex(-1.0, 0.0, -1.0),
                      new Vertex( 1.0, 0.0, -3.0));

      // Two Color objects.
      model.addColor(new Color(255, 127,   0),
                     new Color(  0, 127, 255));

      // Single LineSegment diagonal to the camera's near plane.
      model.addPrimitive(new LineSegment(0, 1, 0, 1));

      this.showCamera = true;
      this.perspective = false;
      System.out.println(model);

      // Create a FrameBufferPanel that will hold a FrameBuffer.
      final int width  = 300;
      final int height = 300;
      fbp = new FrameBufferPanel(width, height, Color.black);

      // Create a JFrame that holds the FrameBufferPanel.
      jf = new JFrame("Renderer 6 - Test NearPlane Clipping v2");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setResizable(false);
      jf.setVisible(true);

      // Register this object as the event listener for JFrame events.
      jf.addKeyListener(this);

      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      fbp.repaint();

      print_help_message();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c && e.isAltDown())
      {
         System.out.println();
         System.out.println(scene.getPosition(0).getModel());
      }
      else if ('d' == c)
      {
         scene.debug = ! scene.debug;
         Clip.debug = scene.debug;
      }
      else if ('D' == c)
      {
         Rasterize.debug = ! Rasterize.debug;
      }
      else if ('a' == c)
      {
         Rasterize.doAntiAliasing = ! Rasterize.doAntiAliasing;
         System.out.print("Anti-aliasing is turned ");
         System.out.println(Rasterize.doAntiAliasing ? "On" : "Off");
      }
      else if ('g' == c)
      {
         Rasterize.doGamma = ! Rasterize.doGamma;
         System.out.print("Gamma correction is turned ");
         System.out.println(Rasterize.doGamma ? "On" : "Off");
      }
      else if ('p' == c)
      {
         perspective = ! perspective;
         final String p = perspective ? "perspective" : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('M' == c)
      {
         showCamera = ! showCamera;
      }
      else if ('n' == c)
      {
         // Move the near plane closer to the camera.
         near -= 0.01;
      }
      else if ('N' == c)
      {
         // Move the near plane away from the camera.
         near += 0.01;
      }
      else if ('b' == c)
      {
         NearClip.doNearClipping = ! NearClip.doNearClipping;
         System.out.print("Near-plane clipping is turned ");
         System.out.println(NearClip.doNearClipping ? "On" : "Off");
      }

      // Set up the camera's view volume.
      final Camera camera;
      if (perspective)
      {
         camera = Camera.projPerspective();
      }
      else
      {
         camera = Camera.projOrtho();
      }
      scene = scene.changeCamera( camera.changeNear(near) );

      if (showCamera && ('M'==c||'n'==c||'N'==c||'p'==c))
      {
         System.out.println( scene.camera );
      }

      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      fbp.repaint();
   }


   private void print_help_message()
   {
      System.out.println("Use the 'd/D' keys to toggle debugging information on and off.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the 'M' key to toggle showing the Camera data.");
      System.out.println("Use the n/N keys to move the camera's near plane.");
      System.out.println("Use the 'b' key to toggle near plane clipping on and off.");
      System.out.println("Use the 'a' key to toggle anti-aliasing on and off.");
      System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new TestNearPlaneClipping_v2()
      );
   }
}
