/*
 * Renderer 6. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;
import renderer.scene.util.CheckModel;
import renderer.framebuffer.*;
import static renderer.pipeline.PipelineLogger.*;

import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our fifth rendering pipeline. It adds a near plane
   clipping stage, {@link NearClip}, before the {@link Projection}
   stage. The near plane clipping stage clips off the part of a
   {@link renderer.scene.primitives.LineSegment} that extends behind
   the {@link Camera}'s near plane. There are five pipeline stages.
*/
public final class Pipeline
{
   // Mostly for compatibility with renderers 1 through 3.
   public static Color DEFAULT_COLOR = Color.white;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      PipelineLogger.debugScene = scene.debug;

      logMessage("\n== Begin Rendering of Scene: " + scene.name + " ==");

      logMessage("-- Current Camera:\n" + scene.camera);

      // For every Position in the Scene, render the Position's Model.
      for (final Position position : scene.positionList)
      {
         PipelineLogger.debugPosition = position.debug;

         if ( position.visible )
         {
            logMessage("==== Render position: " + position.name + " ====");

            logMessage("---- Translation vector = " + position.getTranslation());

            if ( position.getModel().visible )
            {
               logMessage("====== Render model: "
                                  + position.getModel().name + " ======");

               CheckModel.check(position.getModel());

               // Mostly for compatibility with renderers 1 through 3.
               if (  position.getModel().colorList.isEmpty()
                 && !position.getModel().vertexList.isEmpty())
               {
                  for (int i = 0; i < position.getModel().vertexList.size(); ++i)
                  {
                     position.getModel().addColor( DEFAULT_COLOR );
                  }
                  System.err.println("***WARNING: Added default color to model: "
                                    + position.getModel().name + ".");
               }

               logVertexList("0. Model        ", position.getModel());

               // 1. Apply the Position's model-to-camera coordinate transformation.
               final Model model1 = Model2Camera.model2camera(position);

               logVertexList("1. Camera       ", model1);
               logColorList("1. Camera       ", model1);
               logPrimitiveList("1. Camera       ", model1);

               // 2. Clip primitives to the camera's near plane.
               final Model model2 = NearClip.clip(model1,
                                                  scene.camera);

               logVertexList("2. Near_Clipped", model2);
               logColorList("2. Near_Clipped", model2);
               logPrimitiveList("2. Near_Clipped", model2);

               // 3. Apply the Camera's projection transformation.
               final Model model3 = Projection.project(model2,
                                                       scene.camera);

               logVertexList("3. Projected   ", model3);

               // 4. Clip primitives to the camera's view rectangle.
               final Model model4 = Clip.clip(model3);

               logVertexList("4. Clipped     ", model4);
               logColorList("4. Clipped     ", model4);
               logPrimitiveList("4. Clipped     ", model4);

               // 5. Rasterize every visible primitive into pixels.
               Rasterize.rasterize(model4, vp);

               logMessage("====== End model: "
                                  + position.getModel().name + " ======");
            }
            else
            {
               logMessage("====== Hidden model: "
                                  + position.getModel().name + " ======");
            }

            logMessage("==== End position: " + position.name + " ====");
         }
         else
         {
            logMessage("==== Hidden position: " + position.name + " ====");
         }
      }
      logMessage("== End Rendering of Scene ==");
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Pipeline() {
      throw new AssertionError();
   }
}
