/*
 * Renderer 6. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.scene;

/**
   This {@code Camera} data structure represents a camera
   located at the origin, looking down the negative z-axis
   with a near clipping plane.
<p>
   This {@code Camera} has associated to it a "view volume"
   that determines what part of space the camera "sees" when
   we use the camera to take a picture (that is, when we
   render a {@link Scene}).
<p>
   This {@code Camera} can "take a picture" two ways, using
   a perspective projection or a parallel (orthographic)
   projection. Each way of taking a picture has a different
   shape for its view volume.
<p>
   For the perspective projection, the view volume is the
   infinitely long frustum that is formed by cutting at the
   near clipping plane, {@code z = -near}, the infinitely
   long pyramid with its apex at the origin and its cross
   section in the image plane, {@code z = -1}, with edges
   {@code x = -1}, {@code x = +1}, {@code y = -1}, and
   {@code y = +1}.
<p>
   For the orthographic projection, the view volume is the
   semi-infinite rectangular cylinder parallel to the z-axis
   and with base in the near clipping plane, {@code z = -near},
   and edges {@code x = -1}, {@code x = +1}, {@code y = -1},
   and {@code y = +1} (a semi-infinite parallelepiped).
<p>
   When the graphics rendering {@link renderer.pipeline.Pipeline}
   uses this {@code Camera} to render a {@link Scene}, the renderer
   only "sees" the geometry from the scene that is contained in this
   camera's view volume. (Notice that this means the orthographic
   camera can see geometry that is behind the camera. In fact, the
   perspective camera can also sees geometry that is behind the
   camera.) The renderer's {@link renderer.pipeline.NearClip} and
   {@link renderer.pipeline.Clip} pipeline stages are responsible
   for making sure that the scene's geometry that is outside of
   this camera's view volume is not visible.
<p>
   The plane {@code z = -1} is the camera's "image plane". The
   rectangle in the image plane with corners {@code (-1, -1, -1)}
   and {@code (+1, +1, -1)} is the camera's "view rectangle". The
   view rectangle is like the film in a real camera, it is where
   the camera's image appears when you take a picture. The contents
   of the camera's view rectangle is what gets rasterized, by the
   renderer's {@link renderer.pipeline.Rasterize} pipeline stage,
   into a {@link renderer.framebuffer.FrameBuffer}'s
   {@link renderer.framebuffer.FrameBuffer.Viewport}.
<p>
   For both the perspective and the parallel projections, the camera's
   near plane is there to prevent the camera from seeing what is "behind"
   the near plane. For the perspective projection, the near plane also
   prevents the renderer from incorrectly rasterizing line segments that
   cross the camera plane, {@code z = 0}.
*/
public final class Camera
{
   // Choose either perspective or parallel projection.
   public final boolean perspective;
   // Define the view volume's clipping plane.
   public final double n; // near clipping plane

   /**
      A private {@code Camera} constructor for
      use by the static factory methods.
   */
   private Camera(final boolean perspective,
                  final double n)
   {
      this.perspective = perspective;
      this.n = n;
   }


   /**
      This is a static factory method.
      <p>
      Set up this {@code Camera}'s view volume as a perspective projection
      of an infinite view frustum extending along the negative z-axis
      beginning at {@code z = -0.1}.

      @return a new {@code Camera} object with the default perspective parameters
   */
   public static Camera projPerspective()
   {
      return new Camera(true,
                        -0.1);  // near clipping plane (near = +0.1)
   }


   /**
      This is a static factory method.
      <p>
      Set up this {@code Camera}'s view volume as a parallel (orthographic)
      projection of a semi-infinite view parallelepiped extending along the
      z-axis beginning at {@code z = 1} and extending to negative infinity.

      @return a new {@code Camera} object with the default orthographic parameters
   */
   public static Camera projOrtho()
   {
      return new Camera(false,
                        +1.0);  // near clipping plane (near = -1.0)
   }


   /**
      Create a new {@code Camera} that is essentially the same as this
      {@code Camera} but with the given distance from the camera to
      the near clipping plane.
      <p>
      When {@code near} is positive, the near clipping plane is in
      front of the camera. When {@code near} is negative, the near
      clipping plane is behind the camera.

      @param near  distance from the new {@code Camera} to its near clipping plane
      @return a new {@code Camera} object with the given value for near
   */
   public Camera changeNear(final double near)
   {
      return new Camera(this.perspective,
                        -near);
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Camera} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "Camera: \n";
      result += "  perspective = " + perspective + "\n";
      result += "  near = " + -n;
      return result;
   }
}
